# 连接脚本

刚开始的连接脚本为了实现高半核，参考了一下 `rcore`，但是后来在使用的过程中发现参考 `rcore` 的脚本并不能完美的解决我们的问题，刚开始我们的映射地址如下：

```plain
0xffffffff_c0000000 -> 0x80000000 (1G)
```

后来我们开始修改自己的连接脚本和映射地址以及启动代码，启动代码如下：

```rust
const STACK_SIZE: usize = 0x10000;

#[link_section = ".bss.stack"]
static mut STACK: [u8; STACK_SIZE] = [0u8; STACK_SIZE];

#[link_section = ".data.prepage"]
static mut PAGE_TABLE: [PTE; PAGE_ITEM_COUNT] = {
    let mut arr: [PTE; PAGE_ITEM_COUNT] = [PTE::new(); PAGE_ITEM_COUNT];
    // 初始化页表信息
    // 0x00000000_80000000 -> 0x80000000 (1G)
    // 高半核
    // 0xffffffc0_00000000 -> 0x00000000 (1G)
    // 0xffffffc0_80000000 -> 0x80000000 (1G)
    arr[2] = PTE::from_addr(0x8000_0000, PTEFlags::VRWX);
    arr[0x100] = PTE::from_addr(0x0, PTEFlags::GVRWX);
    arr[0x102] = PTE::from_addr(0x8000_0000, PTEFlags::GVRWX);
    arr
};

core::arch::asm!(
    // 1. 设置栈信息
    // sp = bootstack + (hartid + 1) * 0x10000
    "
        la      sp, {boot_stack}
        li      t0, {stack_size}
        add     sp, sp, t0              // set boot stack

        li      s0, {virt_addr_start}   // add virtual address
        add     sp, sp, s0
    ",
    // 2. 开启分页模式
    // satp = (8 << 60) | PPN(page_table)
    "
        la      t0, {page_table}
        srli    t0, t0, 12
        li      t1, 8 << 60
        or      t0, t0, t1
        csrw    satp, t0
        sfence.vma
    ",
    // 3. 跳到 rust_main 函数，绝对路径
    "   
        la      a2, rust_main
        add     a2, a2, s0
        jalr    a2                      // call rust_main
    ",
    stack_size = const STACK_SIZE,
    boot_stack = sym STACK,
    page_table = sym PAGE_TABLE,
    virt_addr_start = const VIRT_ADDR_START,
    options(noreturn),
)
```

但是在设置链接脚本的过程中也产生了一些问题，就是我们发现设置后 `page_table` 并不是 `4k` 对齐的，在连接脚本内设置的 `. = ALIGN(4K)` 是对 `VMA` 进行对齐，而不是 `LMA(L 指 Load)`，因此我们开始寻找新的解决方案，发现在 `section` 后直接跟 `4K` 可以解决这个问题。因此我们的链接脚本改为如下：

```ld
OUTPUT_ARCH(riscv)
ENTRY(_start)

BASE_ADDRESS = 0xffffffc080200000;

__ENTRY_ADDR = 0x80200000;

SECTIONS
{
    /* Load the kernel at this address: "." means the current address */
    . = BASE_ADDRESS;
    start = .;

    .text ALIGN(4K): AT(__ENTRY_ADDR) {
        stext = .;
        *(.text.entry)
        *(.text .text.*)
        . = ALIGN(4K);
        etext = .;
    }

    .rodata ALIGN(4K): {
        srodata = .;
        *(.rodata .rodata.*)
        . = ALIGN(4K);
        erodata = .;
    }

    .data ALIGN(4K): {
        . = ALIGN(4K);
        *(.data.prepage)
        . = ALIGN(4K);
        sdata = .;
        *(.data .data.*)
        *(.sdata .sdata.*)
        edata = .;
    }

    .stack : {
        *(.bss.stack)
    }

    .bss : {
        *(.bss.stack)
        sbss = .;
        *(.bss .bss.*)
        *(.sbss .sbss.*)
        ebss = .;
    }

    PROVIDE(end = .);
}
```

# 下一步

然后我们就可以把下面的地址分配给堆和栈，从而动态的分配内存。目前还需要考虑的问题是，如果堆由于满了再去申请内存导致堆移除，这个时候进行缺页映射会不会再去堆上申请数据？如果申请数据的话大概率还是报错，所以应该怎么解决这个问题？
